"""Library for using the Google Nest SDM API.

See https://developers.google.com/nest/device-access/api for the documentation
on how to use the API.

The primary components in this library are:
- `auth`: You need to implement `AbstractAuth` to provide credentials.
- `google_nest_subscriber`: A wrapper around the pub/sub system for efficiently
  listening to changes in device state.
- `device_manager`: Holds local state for devices, populated by the subscriber.
- `device`: Holds device traits and current device state
- `event_media`: For media related to camera or doorbell events.

Example usage:
```
    subscriber = GoogleNestSubscriber(
        auth_impl, # Your credential provider
        # Follow nest developer API docs to obtain these
        DEVICE_ACCESS_PROJECT_ID,
        SUBSCRIBER_ID,
    )
    await subscriber.start_async()
    device_manager = await subscriber.async_get_device_manager()

    for device in device_manager.devices.values():
        if device.temperature:
            temp = device.temperatureambient_temperature_celsius
            print("Device temperature: {temp:0.2f}")
```
"""

__all__ = [
    "google_nest_subscriber",
    "device_manager",
    "device",
    "camera_traits",
    "device_traits",
    "doorbell_traits",
    "thermostat_traits",
    "structure",
    "auth",
    "event_media",
    "event",
    "exceptions",
    "diagnostics",
]


"""Libraries related to providing a device level interface for event related media.

An `EventMediaManager` is associated with a single device and manages the
state for events and the lifecycle of media for those events. The manager is
invoked by the subscriber when new events arrive and it handles any fetching
related to the media, as well as transcoding video clips of needed. The
`CachePolicy` settings determine lifecycle options such as how many events
to keep around in the underlying store.
"""

from __future__ import annotations

import asyncio
import datetime
import itertools
import logging
import time
from abc import ABC, abstractmethod
from collections import OrderedDict
from collections.abc import Iterable
from dataclasses import dataclass, field
from typing import Any, Awaitable, Callable, cast

try:
    from pydantic.v1 import BaseModel, validator, Field
except ImportError:
    from pydantic import BaseModel, validator, Field  # type: ignore

from .camera_traits import (
    CameraClipPreviewTrait,
    CameraEventImageTrait,
    EventImageContentType,
)
from .diagnostics import EVENT_MEDIA_DIAGNOSTICS as DIAGNOSTICS
from .diagnostics import Diagnostics
from .event import (
    CameraClipPreviewEvent,
    CameraMotionEvent,
    CameraPersonEvent,
    CameraSoundEvent,
    DoorbellChimeEvent,
    EventImageType,
    EventMessage,
    EventToken,
    ImageEventBase,
    session_event_image_type,
)
from .exceptions import GoogleNestException, TranscodeException
from .transcoder import Transcoder

__all__ = [
    "EventMediaManager",
    "ImageSession",
    "ClipPreviewSession",
    "Media",
    "EventMediaStore",
    "CachePolicy",
]

_LOGGER = logging.getLogger(__name__)

# Should be large enough for processing, but not too large to be a size issue
SNAPSHOT_WIDTH_PX = 1600

DEFAULT_CACHE_SIZE = 2

# Events are collapsed in the order shown here
VISIBLE_EVENTS = [
    DoorbellChimeEvent.NAME,
    CameraPersonEvent.NAME,
    CameraMotionEvent.NAME,
    CameraSoundEvent.NAME,
]

# Percentage of items to delete when bulk purging from the cache
EXPIRE_CACHE_BATCH_SIZE = 0.05


@dataclass
class Media:
    """Represents media related to an event."""

    contents: bytes
    """Media content."""

    event_image_type: EventImageContentType
    """Content event image type of the media."""

    @property
    def content_type(self) -> str:
        """Content type of the media."""
        return self.event_image_type.content_type


@dataclass
class ImageSession:
    """An object that holds an image based event."""

    event_token: str
    """A token that can be used to fetch the media for the event."""

    timestamp: datetime.datetime
    """Timestamp when the event happened."""

    event_type: str
    """A label for the type of event."""


@dataclass
class ClipPreviewSession:
    """An object that holds a clip based event."""

    event_token: str
    """A token that can be used to fetch the media for the event."""

    timestamp: datetime.datetime
    """Timestamp when the event happened."""

    event_types: list[str]
    """A label for the type of event."""


class EventMediaStore(ABC):
    """Interface for external storage."""

    @abstractmethod
    async def async_load(self) -> dict | None:
        """Load data."""

    @abstractmethod
    async def async_save(self, data: dict) -> None:
        """Save data."""

    @abstractmethod
    def get_media_key(self, device_id: str, event: ImageEventBase) -> str:
        """Return the filename to use for the device and event."""

    def get_image_media_key(self, device_id: str, event: ImageEventBase) -> str:
        """Return the filename for image media."""
        return self.get_media_key(device_id, event)

    def get_clip_preview_media_key(self, device_id: str, event: ImageEventBase) -> str:
        """Return the filename for clip preview media."""
        return self.get_media_key(device_id, event)

    def get_clip_preview_thumbnail_media_key(
        self, device_id: str, event: ImageEventBase
    ) -> str | None:
        """Return the filename for thumbnail for clip preview media."""
        return None

    async def async_load_media(self, media_key: str) -> bytes | None:
        """Load media content."""

    async def async_save_media(self, media_key: str, content: bytes) -> None:
        """Write media content."""

    async def async_remove_media(self, media_key: str) -> None:
        """Remove media content."""


class InMemoryEventMediaStore(EventMediaStore):
    """An in memory implementation of EventMediaStore."""

    def __init__(self) -> None:
        self._data: dict | None = None
        self._media: dict[str, bytes] = {}

    async def async_load(self) -> dict | None:
        """Load data."""
        return self._data

    async def async_save(self, data: dict) -> None:
        """Save data."""
        self._data = data

    def get_media_key(self, device_id: str, event: ImageEventBase) -> str:
        """Return the media key to use for the device and event."""
        suffix = "jpg" if event.event_image_type == EventImageType.IMAGE else "mp4"
        return f"{device_id}_{event.timestamp}_{event.event_session_id}.{suffix}"

    def get_image_media_key(self, device_id: str, event: ImageEventBase) -> str:
        """Return the media key to use for the device and event."""
        return (
            f"{device_id}_{event.timestamp}_{event.event_session_id}_"
            f"{event.event_id}.jpg"
        )

    def get_clip_preview_media_key(self, device_id: str, event: ImageEventBase) -> str:
        """Return the media key to use for the device and event."""
        return f"{device_id}_{event.timestamp}_{event.event_session_id}.mp4"

    def get_clip_preview_thumbnail_media_key(
        self, device_id: str, event: ImageEventBase
    ) -> str | None:
        """Return the media key to use for the clip preview thumbnail."""
        return f"{device_id}_{event.timestamp}_{event.event_session_id}_thumb.jpg"

    async def async_load_media(self, media_key: str) -> bytes | None:
        """Load media content."""
        with DIAGNOSTICS.timer("load_media"):
            return self._media.get(media_key)

    async def async_save_media(self, media_key: str, content: bytes) -> None:
        """Remove media content."""
        with DIAGNOSTICS.timer("save_media"):
            self._media[media_key] = content

    async def async_remove_media(self, media_key: str) -> None:
        """Remove media content."""
        with DIAGNOSTICS.timer("remove_media"):
            if media_key in self._media:
                del self._media[media_key]


@dataclass
class CachePolicy:
    """Policy for how many local objects to cache in memory."""

    event_cache_size: int = DEFAULT_CACHE_SIZE
    """Number of events to keep in memory per device."""

    fetch: bool = False
    """Determine if event media should be pre-fetched."""

    store: EventMediaStore = field(default_factory=InMemoryEventMediaStore)
    """The EventMediaStore object for storing media content."""

    transcoder: Transcoder | None = None
    """The transcoder for encoding media."""

    @property
    def event_cache_expire_count(self) -> int:
        """Number of events to keep in memory per device."""
        return max(int(self.event_cache_size * EXPIRE_CACHE_BATCH_SIZE), 1)


class EventMediaModelItem(BaseModel):
    """Structure used to persist the event in EventMediaStore."""

    event_session_id: str
    events: dict[str, ImageEventBase] = Field(default_factory=dict)
    media_key: str | None
    event_media_keys: dict[str, str] = Field(default_factory=dict)
    thumbnail_media_key: str | None
    pending_event_keys: set[str] = Field(default_factory=set)

    @staticmethod
    def from_dict(data: dict[str, Any]) -> EventMediaModelItem:
        """Read from serialized dictionary."""
        return EventMediaModelItem(**data)

    @validator("events", pre=True)
    def _validate_events(
        cls, data: dict[str, ImageEventBase | dict] | None
    ) -> dict[str, ImageEventBase]:
        events: dict[str, ImageEventBase] = {}
        for event_type, event_data in (data or {}).items():
            if isinstance(event_data, ImageEventBase):
                events[event_type] = event_data
            elif event := ImageEventBase.from_dict(event_data):
                events[event_type] = event

        # Link events to other events in the session
        event_image_type = session_event_image_type(events.values())
        for event in events.values():
            event.event_image_type = event_image_type
        return events

    @property
    def visible_event(self) -> ImageEventBase | None:
        """Get the primary visible event for this item."""
        for event_type in VISIBLE_EVENTS:
            if event := self.events.get(event_type):
                return event
        return None

    def merge_events(self, new_events: dict[str, ImageEventBase]) -> None:
        """Merge new incoming events with the existing set."""
        new_keys = new_events.keys() - self.events.keys()
        self.events.update(new_events)
        self.pending_event_keys |= new_keys

    @property
    def pending_events(self) -> dict[str, ImageEventBase]:
        """Return all associated events with this record."""
        return {
            key: value
            for key, value in self.events.items()
            if key in self.pending_event_keys
        }

    def notified(self, event_keys: Iterable[str]) -> None:
        """Mark the specified events as notified."""
        self.pending_event_keys = self.pending_event_keys - set(event_keys)

    def media_key_for_token(self, token: EventToken) -> str | None:
        """Return media key for the specified event token."""
        if token.event_id:
            if token.event_id in self.event_media_keys:
                return self.event_media_keys[token.event_id]
            # Fallback to legacy single event per session
        return self.media_key

    @property
    def any_media_key(self) -> str | None:
        """Return any media item for compatibility with legacy APIs."""
        if self.media_key:
            return self.media_key
        if self.event_media_keys.values():
            return next(iter(self.event_media_keys.values()))
        return None

    @property
    def all_media_keys(self) -> list[str]:
        """Return all media items for purging media keys."""
        keys = [self.media_key, self.thumbnail_media_key] + list(
            self.event_media_keys.values()
        )
        return [key for key in keys if key is not None]

    def as_dict(self) -> dict[str, Any]:
        """Return a EventMediaModelItem as a serializable dict."""
        result = self.dict(exclude_unset=True)
        # Overwrite with ImageEventBase custom serialization
        result["events"] = dict((k, v.as_dict()) for k, v in self.events.items())
        return result


class EventMediaManager:
    """Responsible for handling recent events and fetching associated media."""

    def __init__(
        self,
        device_id: str,
        traits: dict[str, Any],
        event_traits: set[str],
        diagnostics: Diagnostics,
    ) -> None:
        """Initialize DeviceEventMediaManager."""
        self._device_id = device_id
        self._traits = traits
        self._event_traits = event_traits
        self._cache_policy = CachePolicy()
        self._callback: Callable[[EventMessage], Awaitable[None]] | None = None
        self._support_fetch = (
            CameraClipPreviewTrait.NAME in traits
            or CameraEventImageTrait.NAME in traits
        )
        self._diagnostics = diagnostics
        self._lock: asyncio.Lock | None = None

    @property
    def cache_policy(self) -> CachePolicy:
        """Return the current CachePolicy."""
        return self._cache_policy

    @cache_policy.setter
    def cache_policy(self, value: CachePolicy) -> None:
        """Update the CachePolicy."""
        self._cache_policy = value

    async def _async_load(self) -> OrderedDict[str, EventMediaModelItem]:
        """Load the device specific data from the store."""
        store_data = await self._cache_policy.store.async_load()
        event_data: OrderedDict[str, EventMediaModelItem] = OrderedDict()
        if store_data:
            device_data = store_data.get(self._device_id, [])
            for item_data in device_data:
                item = EventMediaModelItem.from_dict(item_data)
                event_data[item.event_session_id] = item
        return event_data

    async def _async_update(self, event_data: dict[str, EventMediaModelItem]) -> None:
        """Save the device specific model to the store."""
        # Event order is preserved so popping from the oldest entry works
        device_data: list[dict[str, Any]] = []
        for item in event_data.values():
            device_data.append(item.as_dict())

        # Read data from the store and update information for this device
        store_data = await self._cache_policy.store.async_load()
        if not store_data:
            store_data = {}
        store_data[self._device_id] = device_data
        await self._cache_policy.store.async_save(store_data)

    async def _async_load_item(
        self, event_session_id: str
    ) -> EventMediaModelItem | None:
        """Load the specific item from the store."""
        event_data = await self._async_load()
        return event_data.get(event_session_id)

    async def _async_update_item(self, item: EventMediaModelItem) -> None:
        """Update the specific item in the store."""
        if not self._lock:
            self._lock = asyncio.Lock()
        async with self._lock:
            event_data = await self._async_load()
            event_data[item.event_session_id] = item
            await self._async_update(event_data)

    async def _expire_cache(self) -> None:
        """Garbage collect any items from the cache."""
        if not self._lock:
            self._lock = asyncio.Lock()
        async with self._lock:
            event_data = await self._async_load()
            _LOGGER.debug("Checking cache size %s", len(event_data))
            if len(event_data) <= self._cache_policy.event_cache_size:
                return
            _LOGGER.debug(
                "Expiring cache %s", self._cache_policy.event_cache_expire_count
            )
            # Bulk pop items
            for i in range(0, self._cache_policy.event_cache_expire_count):
                (key, old_item) = event_data.popitem(last=False)
                _LOGGER.debug(
                    "Expiring media %s (%s)",
                    old_item.all_media_keys,
                    old_item.event_session_id,
                )
                for media_key in old_item.all_media_keys:
                    await self._cache_policy.store.async_remove_media(media_key)
            await self._async_update(event_data)

    async def _fetch_media(self, item: EventMediaModelItem) -> None:
        """Fetch media from the server in response to a pubsub event."""
        store = self._cache_policy.store
        if CameraClipPreviewTrait.NAME in self._traits:
            self._diagnostics.increment("fetch_clip")
            if (
                item.media_key
                or not item.visible_event
                or not (
                    clip_event := cast(
                        CameraClipPreviewEvent,
                        item.events.get(CameraClipPreviewEvent.NAME),
                    )
                )
                or clip_event.is_expired
            ):
                self._diagnostics.increment("fetch_clip.skip")
                return
            clip_preview_trait: CameraClipPreviewTrait = self._traits[
                CameraClipPreviewTrait.NAME
            ]
            event_image = await clip_preview_trait.generate_event_image(
                clip_event.preview_url
            )
            if event_image:
                content = await event_image.contents()
                # Caller will persist the media key assignment
                media_key = store.get_clip_preview_media_key(
                    self._device_id, item.visible_event
                )
                item.media_key = media_key
                _LOGGER.debug("Saving media %s (%s)", media_key, item.event_session_id)
                self._diagnostics.increment("fetch_clip.save")
                await store.async_save_media(media_key, content)
                return

        if CameraEventImageTrait.NAME not in self._traits:
            return
        self._diagnostics.increment("fetch_image")

        event_image_trait: CameraEventImageTrait = self._traits[
            CameraEventImageTrait.NAME
        ]
        for event in item.events.values():
            if event.event_id in item.event_media_keys or event.is_expired:
                self._diagnostics.increment("fetch_image.skip")
                continue
            event_image = await event_image_trait.generate_image(event.event_id)
            content = await event_image.contents(width=SNAPSHOT_WIDTH_PX)

            # Caller will persist the media key assignment
            media_key = store.get_image_media_key(self._device_id, event)
            item.event_media_keys[event.event_id] = media_key
            _LOGGER.debug("Saving media %s (%s)", media_key, item.event_session_id)
            self._diagnostics.increment("fetch_image.save")
            await store.async_save_media(media_key, content)

    async def get_media_from_token(self, event_token: str) -> Media | None:
        """Get media based on the event token."""
        token = EventToken.decode(event_token)
        if not (item := await self._async_load_item(token.event_session_id)):
            self._diagnostics.increment("get_media.invalid_event")
            _LOGGER.debug(
                "No event information found for event id: %s", token.event_session_id
            )
            return None
        media_key = item.media_key_for_token(token)
        if not media_key:
            self._diagnostics.increment("get_media.no_media")
            _LOGGER.debug("No persisted media for event id %s", token)
            return None
        contents = await self._cache_policy.store.async_load_media(media_key)
        if not contents:
            self._diagnostics.increment("get_media.empty")
            _LOGGER.debug(
                "Unable to load persisted media for event id: (%s, %s, %s)",
                token.event_session_id,
                token.event_id,
                item.media_key,
            )
            return None
        assert item.visible_event
        self._diagnostics.increment("get_media.success")
        return Media(contents, item.visible_event.event_image_type)

    async def get_clip_thumbnail_from_token(self, event_token: str) -> Media | None:
        """Get a thumbnail from the event token."""
        self._diagnostics.increment("get_clip")
        token = EventToken.decode(event_token)
        if (
            not (item := await self._async_load_item(token.event_session_id))
            or not item.visible_event
        ):
            self._diagnostics.increment("get_clip.invalid_event")
            _LOGGER.debug(
                "No event information found for event id: %s", token.event_session_id
            )
            return None

        if item.thumbnail_media_key:
            # Load cached thumbnail
            contents = await self._cache_policy.store.async_load_media(
                item.thumbnail_media_key
            )
            if contents:
                self._diagnostics.increment("get_clip.cached")
                return Media(contents, EventImageType.IMAGE_PREVIEW)
            _LOGGER.debug(
                "Thumbnail %s does not exist; transcoding", item.thumbnail_media_key
            )

        # Check for existing primary media
        media_key = item.media_key_for_token(token)
        if not media_key:
            self._diagnostics.increment("get_clip.no_media")
            _LOGGER.debug("No persisted media for event id %s", token)
            return None

        thumbnail_media_key = (
            self._cache_policy.store.get_clip_preview_thumbnail_media_key(
                self._device_id, item.visible_event
            )
        )
        if not self._cache_policy.transcoder or not thumbnail_media_key:
            self._diagnostics.increment("get_clip.no_transcoding")
            _LOGGER.debug("Clip transcoding disabled")
            return None

        try:
            await self._cache_policy.transcoder.transcode_clip(
                media_key, thumbnail_media_key
            )
        except TranscodeException as err:
            self._diagnostics.increment("get_clip.transcode_error")
            _LOGGER.debug("Failure to transcode clip thumbnail: %s", str(err))
            return None

        contents = await self._cache_policy.store.async_load_media(thumbnail_media_key)
        if not contents:
            self._diagnostics.increment("get_clip.load_error")
            _LOGGER.debug(
                "Failed to load transcoded clip: %s", item.thumbnail_media_key
            )
            return None

        item.thumbnail_media_key = thumbnail_media_key
        await self._async_update_item(item)

        self._diagnostics.increment("get_clip.success")
        return Media(contents, EventImageType.IMAGE_PREVIEW)

    async def async_image_sessions(self) -> list[ImageSession]:
        """Return revent events."""
        self._diagnostics.increment("load_image_sessions")

        def _get_events(x: EventMediaModelItem) -> list[ImageEventBase]:
            # Only return events that have successful media fetches
            return [
                y
                for y in x.events.values()
                if x.media_key or y.event_id in x.event_media_keys
            ]

        result = await self._items_with_media()
        events_list = list(map(_get_events, result))
        events: Iterable[ImageEventBase] = itertools.chain(*events_list)

        def _get_session(x: ImageEventBase) -> ImageSession:
            return ImageSession(x.event_token, x.timestamp, x.event_type)

        event_result = list(map(_get_session, events))
        event_result.sort(key=lambda x: x.timestamp, reverse=True)
        return event_result

    async def async_clip_preview_sessions(self) -> list[ClipPreviewSession]:
        """Return revent events for a device that supports clips."""
        self._diagnostics.increment("load_clip_previews")

        def _event_visible(x: ImageEventBase) -> bool:
            return x.event_type in VISIBLE_EVENTS

        def _get_event_session(x: EventMediaModelItem) -> ClipPreviewSession | None:
            assert x.visible_event
            events = list(filter(_event_visible, x.events.values()))
            events.sort(key=lambda x: x.timestamp)
            if not events:
                _LOGGER.debug("Partial event in storage")
                return None
            visible_event = events[0]
            return ClipPreviewSession(
                visible_event.event_token,
                visible_event.timestamp,
                [y.event_type for y in events],
            )

        result = await self._items_with_media()
        clips: Iterable[ClipPreviewSession | None] = iter(
            map(_get_event_session, result)
        )
        valid_clips: list[ClipPreviewSession] = [x for x in clips if x is not None]
        valid_clips.sort(key=lambda x: x.timestamp, reverse=True)
        return valid_clips

    async def _items_with_media(self) -> list[EventMediaModelItem]:
        """Return items in the model that have media for serving."""

        def _filter(x: EventMediaModelItem) -> bool:
            """Return events already fetched or that could be fetched."""
            if x.media_key or x.event_media_keys:
                return True
            return False

        event_data = await self._async_load()
        return list(filter(_filter, event_data.values()))

    def set_update_callback(
        self, target: Callable[[EventMessage], Awaitable[None]]
    ) -> None:
        """Register a callback invoked when new messages are received."""
        self._callback = target

    async def async_handle_events(self, event_message: EventMessage) -> None:
        """Handle the EventMessage."""
        self._diagnostics.increment("event")
        event_sessions: dict[
            str, dict[str, ImageEventBase]
        ] | None = event_message.event_sessions
        if not event_sessions:
            return
        _LOGGER.debug("Event Update %s", event_sessions.keys())
        recv_latency_ms = int((time.time() - event_message.timestamp.timestamp()) * 100)

        # Notify traits to cache most recent event
        pairs = list(event_sessions.items())
        for event_session_id, event_dict in pairs:
            supported = False
            for event_name, event in event_dict.items():
                if not event.is_expired:
                    self._diagnostics.elapsed(event_name, recv_latency_ms)
                else:
                    self._diagnostics.elapsed(f"{event_name}_expired", recv_latency_ms)
                if event_name not in self._event_traits:
                    self._diagnostics.increment(f"event.unsupported.{event_name}")
                    _LOGGER.debug("Unsupported event trait: %s", event_name)
                    continue
                supported = True

            # Skip any entirely unsupported events
            if not supported:
                del event_sessions[event_session_id]

        model_items = []
        failure = False
        for event_session_id, event_dict in event_sessions.items():
            # Track all related events together with the same session
            if model_item := await self._async_load_item(event_session_id):
                self._diagnostics.increment("event.update")
                model_item.merge_events(event_dict)
            else:
                self._diagnostics.increment("event.new")
                # A new event session
                model_item = EventMediaModelItem(
                    event_session_id=event_session_id,
                    events=event_dict,
                    media_key=None,
                    event_media_keys={},
                    thumbnail_media_key=None,
                    pending_event_keys=set(event_dict.keys()),
                )
            model_items.append(model_item)

            if self._support_fetch and self._cache_policy.fetch:
                self._diagnostics.increment("event.fetch")
                try:
                    await self._fetch_media(model_item)
                except GoogleNestException as err:
                    self._diagnostics.increment("event.fetch_error")
                    failure = True
                    _LOGGER.warning(
                        "Failure when pre-fetching event '%s': %s",
                        event.event_session_id,
                        str(err),
                    )

        # Send notifications for any undelivered events that have media.
        pending_events: dict[str, ImageEventBase] = {}
        for model_item in model_items:
            if (
                model_item.any_media_key is None
                and self._support_fetch
                and self._cache_policy.fetch
                and not event_message.is_thread_ended
                and not failure
            ):
                continue
            pending_events.update(model_item.pending_events)

        if pending_events:
            _LOGGER.debug("Message contains notifiable events: %s", pending_events)
            event_message = event_message.with_events(
                pending_events.keys(), pending_events
            )
            if self._callback:
                self._diagnostics.increment("event.notify")
                await self._callback(event_message)
        else:
            _LOGGER.debug("Message did not contain notifiable events")

        for model_item in model_items:
            model_item.notified(pending_events.keys())
            await self._async_update_item(model_item)

        await self._expire_cache()



"""Base model for all nest trait based classes."""

from __future__ import annotations

from typing import Any

try:
    from pydantic.v1 import BaseModel, root_validator
except ImportError:
    from pydantic import BaseModel, root_validator  # type: ignore


TRAITS = "traits"
SDM_PREFIX = "sdm."


class TraitModel(BaseModel):
    """Base model for API objects that are trait based.

    This is meant to be subclasses by the model definitions.
    """

    _EXCLUDE_FIELDS = set(
        {
            "_trait_event_ts",
        }
    )

    def __init__(self, **data: Any):
        """Initialize TraitModel."""
        super().__init__(**data)

    @property
    def traits(self) -> dict[str, Any]:
        """Return a trait mixin on None."""
        return {
            field.alias: getattr(self, field.name)
            for field in self.__fields__.values()
            if getattr(self, field.name) is not None
            and field.alias.startswith(SDM_PREFIX)
        }

    @root_validator(pre=True)
    def _parse_traits(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Parse traits as primary members of this class."""
        if traits := values.get(TRAITS):
            values.update(traits)
        return values

    @property
    def raw_data(self) -> dict[str, Any]:
        """Return raw data for the object."""
        return self.dict(
            by_alias=True,
            exclude=self._EXCLUDE_FIELDS,
            exclude_unset=True,
            exclude_defaults=True,
        )

    class Config:
        extra = "allow"



"""Events from pubsub subscriber."""

from __future__ import annotations

import base64
import binascii
import datetime
import hashlib
import json
import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Iterable, Mapping, Final

try:
    from pydantic.v1 import (
        BaseModel,
        Field,
        root_validator,
        validate_arguments,
        validator,
    )
except ImportError:
    from pydantic import (  # type: ignore
        BaseModel,
        Field,
        root_validator,
        validate_arguments,
        validator,
    )

from .auth import AbstractAuth
from .exceptions import DecodeException
from .registry import Registry

__all__ = [
    "EventMessage",
    "CameraMotionEvent",
    "CameraPersonEvent",
    "CameraSoundEvent",
    "DoorbellChimeEvent",
    "CameraClipPreviewEvent",
    "EventImageType",
    "EventProcessingError",
]

EVENT_ID = "eventId"
EVENT_SESSION_ID = "eventSessionId"
TIMESTAMP = "timestamp"
RESOURCE_UPDATE = "resourceUpdate"
NAME = "name"
TRAITS = "traits"
EVENTS = "events"
PREVIEW_URL = "previewUrl"
ZONES = "zones"
EVENT_THREAD_STATE_ENDED = "ENDED"

# Event images expire 30 seconds after the event is published
EVENT_IMAGE_EXPIRE_SECS = 30

# Camera clip previews don't list an expiration in the API. Lets say 15 minutes
# as an arbitrary number for now.
CAMERA_CLIP_PREVIEW_EXPIRE_SECS = 15 * 60

EVENT_MAP = Registry()

_LOGGER = logging.getLogger(__name__)


class EventProcessingError(Exception):
    """Raised when there was an error handling an event."""


@dataclass
class EventImageContentType:
    """Event image content type."""

    content_type: str

    def __str__(self) -> str:
        """Return a string representation of the event image type."""
        return self.content_type


class EventImageType:
    IMAGE = EventImageContentType("image/jpeg")
    CLIP_PREVIEW = EventImageContentType("video/mp4")
    IMAGE_PREVIEW = EventImageContentType("image/gif")

    @staticmethod
    def from_string(content_type: str) -> EventImageContentType:
        """Parse an EventImageType from a string representation."""
        if content_type == EventImageType.CLIP_PREVIEW.content_type:
            return EventImageType.CLIP_PREVIEW
        elif content_type == EventImageType.IMAGE.content_type:
            return EventImageType.IMAGE
        elif content_type == EventImageType.IMAGE_PREVIEW.content_type:
            return EventImageType.IMAGE_PREVIEW
        else:
            return EventImageContentType(content_type)


@dataclass
class EventToken:
    """Identifier for a unique event."""

    event_session_id: str = Field(alias="eventSessionId")
    event_id: str = Field(alias="eventId")

    def encode(self) -> str:
        """Encode the event token as a serialized string."""
        data = [self.event_session_id, self.event_id]
        b = json.dumps(data).encode("utf-8")
        return base64.b64encode(b).decode("utf-8")

    @staticmethod
    def decode(content: str) -> EventToken:
        """Decode an event token into a class."""
        try:
            s = base64.b64decode(content).decode("utf-8")
        except binascii.Error as err:
            raise DecodeException from err
        data = json.loads(s)
        if not isinstance(data, list) or len(data) != 2:
            raise DecodeException("Unexpected data type: %s", data)
        return EventToken(data[0], data[1])

    def __repr__(self) -> str:
        if not self.event_id:
            return "<EventToken event_session_id" + self.event_session_id + ">"
        return (
            "<EventToken event_session_id"
            + self.event_session_id
            + " event_id="
            + self.event_id
            + ">"
        )


class ImageEventBase(BaseModel, ABC):
    """Base class for all image related event types."""

    event_id: str = Field(alias="eventId")
    """ID used to associate separate messages with a single event."""

    event_session_id: str = Field(alias="eventSessionId")
    """ID used to associate separate messages with a single event."""

    zones: list[str] = Field(default_factory=list)
    """List of zones for the event."""

    timestamp: datetime.datetime
    """Timestamp when the event occurred."""

    event_image_type: EventImageContentType
    """Type of the event."""

    def __init__(self, data: Mapping[str, Any], timestamp: datetime.datetime) -> None:
        """Initialize EventBase."""
        super().__init__(**data, timestamp=timestamp)
        self._data = data

    @property
    def event_token(self) -> str:
        """An identifier of this session / event combination."""
        token = EventToken(self.event_session_id, self.event_id)
        return token.encode()

    @property
    @abstractmethod
    def event_type(self) -> str:
        """The type of event."""

    @property
    def expires_at(self) -> datetime.datetime:
        """Timestamp when the message expires."""
        return self.timestamp + datetime.timedelta(seconds=EVENT_IMAGE_EXPIRE_SECS)

    @property
    def is_expired(self) -> bool:
        """Return true if the event expiration has passed."""
        now = datetime.datetime.now(tz=datetime.timezone.utc)
        return self.expires_at < now

    def as_dict(self) -> dict[str, Any]:
        """Return as a dict form that can be serialized."""
        return {
            "event_type": self.event_type,
            "event_data": self._data,
            "timestamp": self.timestamp.isoformat(),
            "event_image_type": str(self.event_image_type),
        }

    @staticmethod
    def from_dict(data: dict[str, Any]) -> ImageEventBase | None:
        """Parse from a serialized dictionary."""
        event_type = data["event_type"]
        event_data = data["event_data"]
        timestamp = datetime.datetime.fromisoformat(data["timestamp"])
        event = _BuildEvent(event_type, event_data, timestamp)
        if event and "event_image_type" in data:
            event.event_image_type = EventImageType.from_string(
                data["event_image_type"]
            )
        return event

    def __repr__(self) -> str:
        return "<ImageEventBase " + str(self.as_dict()) + ">"

    class Config:
        arbitrary_types_allowed = True
        extra = "allow"


@EVENT_MAP.register()
class CameraMotionEvent(ImageEventBase):
    """Motion has been detected by the camera."""

    NAME: Final = "sdm.devices.events.CameraMotion.Motion"
    event_type: Final = "sdm.devices.events.CameraMotion.Motion"
    event_image_type = EventImageType.IMAGE


@EVENT_MAP.register()
class CameraPersonEvent(ImageEventBase):
    """A person has been detected by the camera."""

    NAME: Final = "sdm.devices.events.CameraPerson.Person"
    event_type: Final = "sdm.devices.events.CameraPerson.Person"
    event_image_type = EventImageType.IMAGE


@EVENT_MAP.register()
class CameraSoundEvent(ImageEventBase):
    """Sound has been detected by the camera."""

    NAME: Final = "sdm.devices.events.CameraSound.Sound"
    event_type: Final = "sdm.devices.events.CameraSound.Sound"
    event_image_type = EventImageType.IMAGE


@EVENT_MAP.register()
class DoorbellChimeEvent(ImageEventBase):
    """The doorbell has been pressed."""

    NAME: Final = "sdm.devices.events.DoorbellChime.Chime"
    event_type: Final = "sdm.devices.events.DoorbellChime.Chime"
    event_image_type = EventImageType.IMAGE


@EVENT_MAP.register()
class CameraClipPreviewEvent(ImageEventBase):
    """A video clip is available for preview, without extra download."""

    NAME: Final = "sdm.devices.events.CameraClipPreview.ClipPreview"
    event_type: Final = "sdm.devices.events.CameraClipPreview.ClipPreview"
    event_image_type = EventImageType.CLIP_PREVIEW

    preview_url: str = Field(alias="previewUrl")
    """A url 10 second frame video file in mp4 format."""

    @root_validator(pre=True)
    def validate_event_id(cls, val: dict[str, Any]) -> dict[str, Any]:
        """Use a URL hash as the event id.

        Since clip preview events already have a url associated with them,
        we don't have an event id for downloading the image.
        """
        if "previewUrl" not in val:
            raise ValueError("missing required field previewUrl")
        url = val["previewUrl"]
        val["eventId"] = hashlib.blake2b(url.encode()).hexdigest()
        return val

    @property
    def expires_at(self) -> datetime.datetime:
        """Event ids do not expire."""
        return self.timestamp + datetime.timedelta(
            seconds=CAMERA_CLIP_PREVIEW_EXPIRE_SECS
        )


class RelationUpdate(BaseModel):
    """Represents a relational update for a resource."""

    type: str
    """Type of relation event 'CREATED', 'UPDATED', 'DELETED'."""

    subject: str
    """Resource that the object is now in relation with."""

    object: str
    """Resource that triggered the event."""


def _BuildEvents(
    events: Mapping[str, Any],
    timestamp: datetime.datetime,
) -> dict[str, ImageEventBase]:
    """Build a trait map out of a response dict."""
    result = {}
    for event_type, event_data in events.items():
        image_event = _BuildEvent(event_type, event_data, timestamp)
        if not image_event:
            continue
        result[event_type] = image_event
    return result


def _BuildEvent(
    event_type: str, event_data: Mapping[str, Any], timestamp: datetime.datetime
) -> ImageEventBase | None:
    if event_type not in EVENT_MAP:
        return None
    cls = EVENT_MAP[event_type]
    return cls(event_data, timestamp)  # type: ignore


def session_event_image_type(events: Iterable[ImageEventBase]) -> EventImageContentType:
    """Determine the event type to use based on the events in the session."""
    for event in events:
        if event.event_image_type != EventImageType.IMAGE:
            return event.event_image_type
    return EventImageType.IMAGE


@validate_arguments
def _validate_datetime(value: datetime.datetime) -> datetime.datetime:
    return value


class EventMessage(BaseModel):
    """Event for a change in trait value or device action."""

    event_id: str = Field(alias="eventId")
    timestamp: datetime.datetime
    resource_update_name: str | None
    resource_update_events: dict[str, ImageEventBase] | None
    resource_update_traits: dict[str, Any] | None
    event_thread_state: str | None = Field(alias="eventThreadState")
    relation_update: RelationUpdate | None = Field(alias="relationUpdate")

    def __init__(self, raw_data: Mapping[str, Any], auth: AbstractAuth) -> None:
        """Initialize an EventMessage."""
        _LOGGER.debug("EventMessage raw_data=%s", raw_data)
        super().__init__(**raw_data, auth=auth)
        self._auth = auth

    @root_validator(pre=True)
    def _parse_resource_update(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Parse resource updates."""
        if update := values.get(RESOURCE_UPDATE):
            if name := update.get(NAME):
                values["resource_update_name"] = name
            if events := update.get(EVENTS):
                values["resource_update_events"] = events
                values["resource_update_events"][TIMESTAMP] = values.get(TIMESTAMP)
            if traits := update.get(TRAITS):
                values["resource_update_traits"] = traits
                values["resource_update_traits"][NAME] = update.get(NAME)
        return values

    @validator("resource_update_events", pre=True)
    def _parse_resource_update_events(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Parse resource updates for events."""
        return _BuildEvents(values, _validate_datetime(values[TIMESTAMP]))

    @property
    def event_sessions(self) -> dict[str, dict[str, ImageEventBase]] | None:
        events = self.resource_update_events
        if not events:
            return None
        event_sessions: dict[str, dict[str, ImageEventBase]] = {}
        for event_name, event in events.items():
            d = event_sessions.get(event.event_session_id, {})
            d[event_name] = event
            event_sessions[event.event_session_id] = d
        # Build associations between all events
        for event_session_id, event_dict in event_sessions.items():
            event_image_type = session_event_image_type(events.values())
            for event_type, event in event_dict.items():
                event.event_image_type = event_image_type
        return event_sessions

    @property
    def raw_data(self) -> dict[str, Any]:
        """Return raw data for the event."""
        return self.dict()

    def with_events(
        self,
        event_keys: Iterable[str],
        merge_data: dict[str, ImageEventBase] | None = None,
    ) -> EventMessage:
        """Create a new EventMessage minus some existing events by key."""
        new_message = self.copy()
        if not merge_data:
            merge_data = {}
        new_events = {}
        for key in event_keys:
            if (
                new_message.resource_update_events
                and key in new_message.resource_update_events
            ):
                new_events[key] = new_message.resource_update_events[key]
            elif merge_data and key in merge_data:
                new_events[key] = merge_data[key]
        new_message.resource_update_events = new_events
        return new_message

    @property
    def is_thread_ended(self) -> bool:
        """Return true if the message indicates the thread is ended."""
        return self.event_thread_state == EVENT_THREAD_STATE_ENDED

    def __repr__(self) -> str:
        """Debug information."""
        return f"EventMessage{self.raw_data}"

    class Config:
        arbitrary_types_allowed = True
        extra = "allow"



"""Subscriber for the Smart Device Management event based API."""
from __future__ import annotations

import asyncio
import concurrent.futures
import enum
import json
import logging
import re
import time
from abc import ABC, abstractmethod
from typing import Awaitable, Callable

from aiohttp.client_exceptions import ClientError
from google.api_core.exceptions import GoogleAPIError, NotFound, Unauthenticated
from google.auth.exceptions import RefreshError, GoogleAuthError, TransportError
from google.auth.transport.requests import Request
from google.cloud import pubsub_v1
from google.oauth2.credentials import Credentials
from google.protobuf.duration_pb2 import Duration

from .auth import AbstractAuth
from .device_manager import DeviceManager
from .diagnostics import SUBSCRIBER_DIAGNOSTICS as DIAGNOSTICS
from .event import EventMessage
from .event_media import CachePolicy
from .exceptions import AuthException, ConfigurationException, SubscriberException
from .google_nest_api import GoogleNestAPI

__all__ = [
    "GoogleNestSubscriber",
    "ApiEnv",
]


_LOGGER = logging.getLogger(__name__)

# Used to catch invalid subscriber id
EXPECTED_SUBSCRIBER_REGEXP = re.compile("projects/.*/subscriptions/.*")

# Used to catch a topic misconfiguration
EXPECTED_TOPIC_REGEXP = re.compile("projects/sdm-[a-z]+/topics/.*")

WATCHDOG_CHECK_INTERVAL_SECONDS = 10

# Restart the subscriber after some delay, with exponential backoff
WATCHDOG_RESTART_DELAY_MIN_SECONDS = 10
WATCHDOG_RESTART_DELAY_MAX_SECONDS = 300
# Reset watchdog backoff
WATCHDOG_RESET_THRESHOLD_SECONDS = 60

DEFAULT_MESSAGE_RETENTION_SECONDS = 15 * 60  # 15 minutes

# Note: Users of non-prod instances will have to manually configure a topic
TOPIC_FORMAT = "projects/sdm-prod/topics/enterprise-{project_id}"

OAUTH2_AUTHORIZE_FORMAT = (
    "https://nestservices.google.com/partnerconnections/{project_id}/auth"
)
OAUTH2_TOKEN = "https://www.googleapis.com/oauth2/v4/token"
SDM_SCOPES = [
    "https://www.googleapis.com/auth/sdm.service",
    "https://www.googleapis.com/auth/pubsub",
]
API_URL = "https://smartdevicemanagement.googleapis.com/v1"


class ApiEnv(enum.Enum):
    PROD = (OAUTH2_AUTHORIZE_FORMAT, API_URL)
    PREPROD = (
        "https://sdmresourcepicker-preprod.sandbox.google.com/partnerconnections/{project_id}/auth",
        "https://preprod-smartdevicemanagement.googleapis.com/v1",
    )

    def __init__(self, authorize_url: str, api_url: str) -> None:
        """Init ApiEnv."""
        self._authorize_url = authorize_url
        self._api_url = api_url

    @property
    def authorize_url_format(self) -> str:
        """OAuth Authorize url format string."""
        return self._authorize_url

    @property
    def api_url(self) -> str:
        """API url."""
        return self._api_url


def get_api_env(env: str | None) -> ApiEnv:
    """Create an ApiEnv from a string."""
    if env is None or env == "prod":
        return ApiEnv.PROD
    if env == "preprod":
        return ApiEnv.PREPROD
    raise ValueError("Invalid ApiEnv: %s" % env)


def _validate_subscription_name(subscription_name: str) -> None:
    """Validates that a subscription name is correct.

    Raises ConfigurationException on failure.
    """
    if not EXPECTED_SUBSCRIBER_REGEXP.match(subscription_name):
        DIAGNOSTICS.increment("subscription_name_invalid")
        raise ConfigurationException(
            "Subscription misconfigured. Expected subscriber_id to "
            f"match '{EXPECTED_SUBSCRIBER_REGEXP.pattern}' but was "
            f"'{subscription_name}'"
        )


def _validate_topic_name(topic_name: str) -> None:
    """Validates that a topic name is correct.

    Raises ConfigurationException on failure.
    """
    if not EXPECTED_TOPIC_REGEXP.match(topic_name):
        DIAGNOSTICS.increment("topic_name_invalid")
        raise ConfigurationException(
            "Subscription misconfigured. Expected topic name to "
            f"match '{EXPECTED_TOPIC_REGEXP.pattern}' but was "
            f"'{topic_name}'."
        )


def refresh_creds(creds: Credentials) -> Credentials:
    """Refresh credentials.

    This is not part of the subscriber API, exposed only to facilitate testing.
    """
    try:
        creds.refresh(Request())
    except RefreshError as err:
        raise AuthException(f"Authentication refresh failure: {err}") from err
    except TransportError as err:
        raise SubscriberException(
            f"Connectivity error during authentication refresh: {err}"
        ) from err
    except GoogleAuthError as err:
        raise SubscriberException(
            f"Error during authentication refresh: {err}"
        ) from err
    return creds


class AbstractSubscriberFactory(ABC):
    """Abstract class for creating a subscriber, to facilitate testing."""

    @abstractmethod
    async def async_create_subscription(
        self,
        creds: Credentials,
        subscription_name: str,
        topic_name: str,
        loop: asyncio.AbstractEventLoop,
    ) -> None:
        """Creates a subscription name if it does not already exist."""

    @abstractmethod
    async def async_delete_subscription(
        self,
        creds: Credentials,
        subscription_name: str,
        loop: asyncio.AbstractEventLoop,
    ) -> None:
        """Deletes a subscription."""

    @abstractmethod
    async def async_new_subscriber(
        self,
        creds: Credentials,
        subscription_name: str,
        loop: asyncio.AbstractEventLoop,
        async_callback: Callable[
            [pubsub_v1.subscriber.message.Message], Awaitable[None]
        ],
    ) -> pubsub_v1.subscriber.futures.StreamingPullFuture:
        """Create a new event subscriber."""


class DefaultSubscriberFactory(AbstractSubscriberFactory):
    """Default implementation that creates Google Pubsub subscriber."""

    async def async_create_subscription(
        self,
        creds: Credentials,
        subscription_name: str,
        topic_name: str,
        loop: asyncio.AbstractEventLoop,
    ) -> None:
        """Creates a subscription name if it does not already exist."""
        _validate_subscription_name(subscription_name)
        _validate_topic_name(topic_name)
        with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor:
            await loop.run_in_executor(
                executor,
                self._create_subscription,
                creds,
                subscription_name,
                topic_name,
            )

    def _create_subscription(
        self,
        creds: Credentials,
        subscription_name: str,
        topic_name: str,
    ) -> None:
        """Creates a subscription name if it does not already exist."""
        subscriber = pubsub_v1.SubscriberClient(credentials=creds)

        subscription = None
        try:
            subscription = subscriber.get_subscription(subscription=subscription_name)
        except NotFound:
            _LOGGER.debug("Existing subscription not found; Creating")
        if subscription:
            if subscription.topic != topic_name:
                raise ConfigurationException(
                    "Subscription misconfigured. Expected topic name to "
                    f"match '{topic_name}' but was "
                    f"'{subscription.topic}'. Please delete in cloud "
                    "console and it will be re-created."
                )
            # Valid subscription already exists; No-op
            return

        message_retention_duration = Duration()
        message_retention_duration.FromSeconds(DEFAULT_MESSAGE_RETENTION_SECONDS)
        subscription_request = {
            "name": subscription_name,
            "topic": topic_name,
            "message_retention_duration": message_retention_duration,
        }
        _LOGGER.debug(f"Creating subscription: {subscription_request}")
        subscriber.create_subscription(request=subscription_request)

    async def async_delete_subscription(
        self,
        creds: Credentials,
        subscription_name: str,
        loop: asyncio.AbstractEventLoop,
    ) -> None:
        """Creates a subscription name if it does not already exist."""
        with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor:
            await loop.run_in_executor(
                executor,
                self._delete_subscription,
                creds,
                subscription_name,
            )

    def _delete_subscription(
        self,
        creds: Credentials,
        subscription_name: str,
    ) -> None:
        """Deletes a subscription."""
        creds = refresh_creds(creds)
        subscriber = pubsub_v1.SubscriberClient(credentials=creds)
        _LOGGER.debug(f"Deleting subscription '{subscription_name}'")
        subscriber.delete_subscription(subscription=subscription_name)

    async def async_new_subscriber(
        self,
        creds: Credentials,
        subscription_name: str,
        loop: asyncio.AbstractEventLoop,
        async_callback: Callable[
            [pubsub_v1.subscriber.message.Message], Awaitable[None]
        ],
    ) -> pubsub_v1.subscriber.futures.StreamingPullFuture:
        """Create a new subscriber with a blocking to async bridge."""

        def callback_wrapper(message: pubsub_v1.subscriber.message.Message) -> None:
            if loop.is_closed():
                return
            future: concurrent.futures.Future = asyncio.run_coroutine_threadsafe(
                async_callback(message), loop
            )
            future.result()

        with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor:
            return await loop.run_in_executor(
                executor,
                self._new_subscriber,
                creds,
                subscription_name,
                callback_wrapper,
            )

    def _new_subscriber(
        self,
        creds: Credentials,
        subscription_name: str,
        callback_wrapper: Callable[[pubsub_v1.subscriber.message.Message], None],
    ) -> pubsub_v1.subscriber.futures.StreamingPullFuture:
        """Issue a command to verify subscriber creds are correct."""
        creds = refresh_creds(creds)
        subscriber = pubsub_v1.SubscriberClient(credentials=creds)
        subscription = subscriber.get_subscription(subscription=subscription_name)
        if subscription.topic:
            _validate_topic_name(subscription.topic)
            _LOGGER.debug(
                "Subscriber '%s' configured on topic '%s'",
                subscription_name,
                subscription.topic,
            )
        return subscriber.subscribe(subscription_name, callback_wrapper)


class GoogleNestSubscriber:
    """Subscribe to events from the Google Nest feed."""

    def __init__(
        self,
        auth: AbstractAuth,
        project_id: str,
        subscriber_id: str,
        subscriber_factory: AbstractSubscriberFactory = DefaultSubscriberFactory(),
        loop: asyncio.AbstractEventLoop | None = None,
        watchdog_check_interval_seconds: float = WATCHDOG_CHECK_INTERVAL_SECONDS,
        watchdog_restart_delay_min_seconds: float = WATCHDOG_RESTART_DELAY_MIN_SECONDS,
    ):
        """Initialize the subscriber for the specified topic."""
        self._auth = auth
        self._subscriber_id = subscriber_id
        self._project_id = project_id
        self._api = GoogleNestAPI(auth, project_id)
        self._loop = loop or asyncio.get_event_loop()
        self._device_manager_task: asyncio.Task[DeviceManager] | None = None
        self._subscriber_factory = subscriber_factory
        self._subscriber_future: pubsub_v1.subscriber.futures.StreamingPullFuture | None = (  # noqa: E501
            None
        )
        self._callback: Callable[[EventMessage], Awaitable[None]] | None = None
        self._healthy = True
        self._watchdog_check_interval_seconds = watchdog_check_interval_seconds
        self._watchdog_restart_delay_min_seconds = watchdog_restart_delay_min_seconds
        self._watchdog_restart_delay_seconds = watchdog_restart_delay_min_seconds
        self._watchdog_task: asyncio.Task | None = None
        self._cache_policy = CachePolicy()

    @property
    def subscriber_id(self) -> str:
        """Return the configured subscriber_id."""
        return self._subscriber_id

    @property
    def project_id(self) -> str:
        """Return the configured SDM project_id."""
        return self._project_id

    def set_update_callback(
        self, target: Callable[[EventMessage], Awaitable[None]]
    ) -> None:
        """Register a callback invoked when new messages are received."""
        self._callback = target
        if self._device_manager_task and self._device_manager_task.done():
            self._device_manager_task.result().set_update_callback(target)

    async def create_subscription(self) -> None:
        """Create the subscription if it does not already exist."""
        _validate_subscription_name(self._subscriber_id)
        DIAGNOSTICS.increment("create_subscription.attempt")
        try:
            creds = await self._auth.async_get_creds()
        except ClientError as err:
            DIAGNOSTICS.increment("create_subscription.creds_error")
            raise AuthException(f"Access token failure: {err}") from err
        try:
            await self._subscriber_factory.async_create_subscription(
                creds,
                self._subscriber_id,
                TOPIC_FORMAT.format(project_id=self._project_id),
                self._loop,
            )
        except NotFound as err:
            DIAGNOSTICS.increment("create_subscription.not_found")
            raise ConfigurationException(
                f"Failed to create subscriber '{self._subscriber_id}' "
                + "(cloud project id incorrect?)"
            ) from err
        except Unauthenticated as err:
            DIAGNOSTICS.increment("create_subscription.unauthenticated")
            raise AuthException("Failed to authenticate subscriber: {err}") from err
        except GoogleAPIError as err:
            DIAGNOSTICS.increment("create_subscription.api_error")
            raise SubscriberException(
                f"Failed to create subscriber '{self._subscriber_id}': {err}"
            ) from err

    async def delete_subscription(self) -> None:
        """Delete the subscription."""
        DIAGNOSTICS.increment("delete_subscription.attempt")
        try:
            creds = await self._auth.async_get_creds()
        except ClientError as err:
            DIAGNOSTICS.increment("delete_subscription.creds_error")
            raise AuthException(f"Access token failure: {err}") from err

        try:
            await self._subscriber_factory.async_delete_subscription(
                creds, self._subscriber_id, self._loop
            )
        except NotFound:
            DIAGNOSTICS.increment("delete_subscription.not_found")
            # No-op if subscription was already deleted
            return
        except Unauthenticated as err:
            DIAGNOSTICS.increment("delete_subscription.unauthenticated")
            raise AuthException("Failed to authenticate subscriber: {err}") from err
        except GoogleAPIError as err:
            DIAGNOSTICS.increment("delete_subscription.api_error")
            raise SubscriberException(
                f"Failed to delete subscription '{self._subscriber_id}': {err}"
            ) from err

    async def start_async(self) -> None:
        """Start the subscriber."""
        DIAGNOSTICS.increment("start")
        _validate_subscription_name(self._subscriber_id)
        try:
            creds = await self._auth.async_get_creds()
        except ClientError as err:
            DIAGNOSTICS.increment("start.creds_error")
            raise AuthException(f"Access token failure: {err}") from err

        try:
            self._subscriber_future = (
                await self._subscriber_factory.async_new_subscriber(
                    creds, self._subscriber_id, self._loop, self._async_message_callback
                )
            )
        except NotFound as err:
            DIAGNOSTICS.increment("start.not_found_error")
            raise ConfigurationException(
                f"Failed to create subscriber '{self._subscriber_id}' id was not found"
            ) from err
        except Unauthenticated as err:
            DIAGNOSTICS.increment("start.unauthenticated")
            raise AuthException("Failed to authenticate subscriber: {err}") from err
        except GoogleAPIError as err:
            DIAGNOSTICS.increment("start.api_error")
            raise SubscriberException(
                f"Failed to create subscriber '{self._subscriber_id}': {err}"
            ) from err

        if not self._healthy:
            _LOGGER.debug("Subscriber reconnected")
            self._healthy = True
            self._watchdog_restart_delay_seconds = (
                self._watchdog_restart_delay_min_seconds
            )
        if self._watchdog_check_interval_seconds > 0:
            self._watchdog_task = asyncio.create_task(self._watchdog())
        if self._subscriber_future:
            self._subscriber_future.add_done_callback(self._done_callback)

    async def _watchdog(self) -> None:
        """Background task that watches the subscriber and restarts it."""
        _LOGGER.debug("Starting background watchdog thread")
        while True:
            if self._subscriber_future and self._subscriber_future.done():
                _LOGGER.debug(
                    "Watchdog: subscriber shut down; restarting in %s seconds",
                    self._watchdog_restart_delay_seconds,
                )
                await asyncio.sleep(self._watchdog_restart_delay_seconds)
                self._watchdog_restart_delay_seconds *= 2
                self._watchdog_restart_delay_seconds = max(
                    self._watchdog_restart_delay_seconds,
                    WATCHDOG_RESTART_DELAY_MAX_SECONDS,
                )
                await self.start_async()
            await asyncio.sleep(self._watchdog_check_interval_seconds)

    def stop_async(self) -> None:
        """Tell the subscriber to start shutting down."""
        DIAGNOSTICS.increment("stop")
        if self._device_manager_task:
            self._device_manager_task.cancel()
        if self._watchdog_task:
            self._watchdog_task.cancel()
        if self._subscriber_future:
            self._subscriber_future.cancel()

    @property
    def cache_policy(self) -> CachePolicy:
        """Return cache policy shared by device EventMediaManager objects."""
        return self._cache_policy

    async def async_get_device_manager(self) -> DeviceManager:
        """Return the DeviceManger with the current state of devices."""
        if not self._device_manager_task:
            self._device_manager_task = asyncio.create_task(
                self._async_create_device_manager()
            )
        assert self._device_manager_task
        return await self._device_manager_task

    async def _async_create_device_manager(self) -> DeviceManager:
        """Create a DeviceManager, populated with initial state."""
        device_manager = DeviceManager(self._cache_policy)
        structures = await self._api.async_get_structures()
        for structure in structures:
            device_manager.add_structure(structure)
        # Subscriber starts after a device fetch
        devices = await self._api.async_get_devices()
        for device in devices:
            device_manager.add_device(device)
        if self._callback:
            device_manager.set_update_callback(self._callback)
        return device_manager

    def _done_callback(
        self, future: pubsub_v1.subscriber.futures.StreamingPullFuture
    ) -> None:
        """Teardown subscriber."""
        if future.exception():
            ex = future.exception()
            if self._healthy:
                self._healthy = False
            _LOGGER.debug("Subscriber disconnected, will restart: %s: %s", type(ex), ex)

    async def _async_message_callback(
        self, message: pubsub_v1.subscriber.message.Message
    ) -> None:
        """Handle a received message."""
        payload = json.loads(bytes.decode(message.data))
        event = EventMessage(payload, self._auth)
        recv = time.time()
        latency_ms = int((recv - event.timestamp.timestamp()) * 1000)
        DIAGNOSTICS.elapsed("message_received", latency_ms)
        # Only accept device events once the Device Manager has been loaded.
        # We are ok with missing messages on startup since the device manager
        # will do a live read. This checks for an exception to avoid throwing
        # inside the pubsub callback and further weding the pubsub client library.
        if (
            self._device_manager_task
            and self._device_manager_task.done()
            and not self._device_manager_task.exception()
        ):
            await self._device_manager_task.result().async_handle_event(event)
            message.ack()

        ack_latency_ms = int((time.time() - recv) * 1000)
        DIAGNOSTICS.elapsed("message_acked", ack_latency_ms)



"""Diagnostics for debugging."""

from __future__ import annotations

import time
from collections import Counter
from collections.abc import Mapping
from contextlib import contextmanager
from typing import Any, Generator, TypeVar, cast

__all__ = [
    "get_diagnostics",
]

class Diagnostics:
    """Information for the library."""

    def __init__(self) -> None:
        """Initialize Diagnostics."""
        self._counter: Counter = Counter()
        self._subkeys: dict[str, Diagnostics] = {}

    def increment(self, key: str, count: int = 1) -> None:
        """Increment a counter for the specified key/event."""
        self._counter.update(Counter({key: count}))

    def elapsed(self, key_prefix: str, elapsed_ms: int = 1) -> None:
        """Track a latency event for the specified key/event prefix."""
        self.increment(f"{key_prefix}_count", 1)
        self.increment(f"{key_prefix}_sum", elapsed_ms)

    def as_dict(self) -> Mapping[str, Any]:
        """Return diagnostics as a debug dictionary."""
        data: dict[str, Any] = {k: self._counter[k] for k in self._counter}
        for k, d in self._subkeys.items():
            v = d.as_dict()
            if not v:
                continue
            data[k] = v
        return data

    def subkey(self, key: str) -> Diagnostics:
        """Return sub-Diagnositics object with the specified subkey."""
        if key not in self._subkeys:
            self._subkeys[key] = Diagnostics()
        return self._subkeys[key]

    @contextmanager
    def timer(self, key_prefix: str) -> Generator[None, None, None]:
        """A context manager that records the timing of operations as a diagnostic."""
        start = time.perf_counter()
        try:
            yield
        finally:
            end = time.perf_counter()
            ms = int((end - start) * 1000)
            self.elapsed(key_prefix, ms)

    def reset(self) -> None:
        """Clear all diagnostics, for testing."""
        self._counter = Counter()
        for d in self._subkeys.values():
            d.reset()


SUBSCRIBER_DIAGNOSTICS = Diagnostics()
EVENT_DIAGNOSTICS = Diagnostics()
EVENT_MEDIA_DIAGNOSTICS = Diagnostics()

MAP = {
    "subscriber": SUBSCRIBER_DIAGNOSTICS,
    "event": EVENT_DIAGNOSTICS,
    "event_media": EVENT_MEDIA_DIAGNOSTICS,
}


def reset() -> None:
    """Clear all diagnostics, for testing."""
    for diagnostics in MAP.values():
        diagnostics.reset()


def get_diagnostics() -> dict[str, Any]:
    """Produce diagnostics information for the library."""
    return {k: v.as_dict() for (k, v) in MAP.items() if v.as_dict()}


REDACT_KEYS = {
    "name",
    "customName",
    "displayName",
    "parent",
    "assignee",
    "subject",
    "object",
    "userId",
    "resourceGroup",
    "eventId",
    "eventSessionId",
    "eventThreadId",
}
REDACTED = "**REDACTED**"


T = TypeVar("T")


def redact_data(data: T) -> T | dict | list:
    """Redact sensitive data in a dict."""
    if not isinstance(data, (Mapping, list)):
        return data

    if isinstance(data, list):
        return cast(T, [redact_data(item) for item in data])

    redacted = {**data}

    for key, value in redacted.items():
        if key in REDACT_KEYS:
            redacted[key] = REDACTED
        elif isinstance(value, dict):
            redacted[key] = redact_data(value)
        elif isinstance(value, list):
            redacted[key] = [redact_data(item) for item in value]

    return redacted



"""Library for traits about devices."""

import datetime
from typing import Any, Dict, Final

try:
    from pydantic.v1 import BaseModel, Field
except ImportError:
    from pydantic import BaseModel, Field  # type: ignore

import aiohttp

from .traits import CommandModel


class ConnectivityTrait(BaseModel):
    """This trait belongs to any device that has connectivity information."""

    NAME: Final = "sdm.devices.traits.Connectivity"

    status: str
    """Device connectivity status.

    Return:
        "OFFLINE", "ONLINE"
    """


class FanTrait(CommandModel):
    """This trait belongs to any device that can control the fan."""

    NAME: Final = "sdm.devices.traits.Fan"

    timer_mode: str | None = Field(alias="timerMode")
    """Timer mode for the fan.

    Return:
        "ON", "OFF"
    """

    timer_timeout: datetime.datetime | None = Field(alias="timerTimeout")

    async def set_timer(
        self, timer_mode: str, duration: int | None = None
    ) -> aiohttp.ClientResponse:
        """Change the fan timer."""
        data: Dict[str, Any] = {
            "command": "sdm.devices.commands.Fan.SetTimer",
            "params": {
                "timerMode": timer_mode,
            },
        }
        if duration:
            data["params"]["duration"] = f"{duration}s"
        return await self.cmd.execute(data)


class InfoTrait(BaseModel):
    """This trait belongs to any device for device-related information."""

    NAME: Final = "sdm.devices.traits.Info"

    custom_name: str | None = Field(alias="customName")
    """Name of the device."""


class HumidityTrait(BaseModel):
    """This trait belongs to any device that has a sensor to measure humidity."""

    NAME: Final = "sdm.devices.traits.Humidity"

    ambient_humidity_percent: float = Field(alias="ambientHumidityPercent")
    """Percent humidity, measured at the device."""


class TemperatureTrait(BaseModel):
    """This trait belongs to any device that has a sensor to measure temperature."""

    NAME: Final = "sdm.devices.traits.Temperature"

    ambient_temperature_celsius: float = Field(alias="ambientTemperatureCelsius")
    """Percent humidity, measured at the device."""



"""Traits belonging to doorbell devices."""

from __future__ import annotations

import logging
from typing import Final

try:
    from pydantic.v1 import BaseModel
except ImportError:
    from pydantic import BaseModel  # type: ignore

from .event import DoorbellChimeEvent

_LOGGER = logging.getLogger(__name__)


class DoorbellChimeTrait(BaseModel):
    """For any device that supports a doorbell chime and related press events."""

    NAME: Final = "sdm.devices.traits.DoorbellChime"
    EVENT_NAME: Final[str] = DoorbellChimeEvent.NAME

    class Config:
        extra = "allow"
        arbitrary_types_allowed = True



"""Traits for thermostats."""

from __future__ import annotations

from abc import ABC
from typing import Final

import aiohttp

try:
    from pydantic.v1 import Field
except ImportError:
    from pydantic import Field  # type: ignore

from .traits import CommandModel
from .model import TraitModel

__all__ = [
    "ThermostatEcoTrait",
    "ThermostatHvacTrait",
    "ThermostatModeTrait",
    "ThermostatTemperatureSetpointTrait",
]

STATUS: Final = "status"
AVAILABLE_MODES: Final = "availableModes"
MODE: Final = "mode"


class ThermostatHeatCoolTrait(CommandModel, ABC):
    """Parent class for traits related to temperature set points."""

    heat_celsius: float | None = Field(alias="heatCelsius")
    """Lowest temperature where thermostat begins heating."""

    cool_celsius: float | None = Field(alias="coolCelsius")
    """Highest cooling temperature where thermostat begins cooling."""


class ThermostatEcoTrait(ThermostatHeatCoolTrait):
    """This trait belongs to any device that has a sensor to measure temperature."""

    NAME: Final = "sdm.devices.traits.ThermostatEco"

    available_modes: list[str] = Field(alias="availableModes", default_factory=list)
    """List of supported Eco modes."""

    mode: str = Field(alias="mode", default="OFF")
    """Eco mode of the thermostat."""

    async def set_mode(self, mode: str) -> aiohttp.ClientResponse:
        """Change the thermostat Eco mode."""
        data = {
            "command": "sdm.devices.commands.ThermostatEco.SetMode",
            "params": {"mode": mode},
        }
        return await self.cmd.execute(data)


class ThermostatHvacTrait(TraitModel):
    """This trait belongs to devices that can report HVAC details."""

    NAME: Final = "sdm.devices.traits.ThermostatHvac"

    status: str = Field("status")
    """HVAC status of the thermostat."""


class ThermostatModeTrait(CommandModel):
    """This trait belongs to devices that support different thermostat modes."""

    NAME: Final = "sdm.devices.traits.ThermostatMode"

    available_modes: list[str] = Field(alias="availableModes")
    """List of supported thermostat modes."""

    mode: str = Field(alias="mode")
    """Mode of the thermostat."""

    async def set_mode(self, mode: str) -> aiohttp.ClientResponse:
        """Change the thermostat Eco mode."""
        data = {
            "command": "sdm.devices.commands.ThermostatMode.SetMode",
            "params": {"mode": mode},
        }
        return await self.cmd.execute(data)


class ThermostatTemperatureSetpointTrait(ThermostatHeatCoolTrait):
    """This trait belongs to devices that support setting target temperature."""

    NAME: Final = "sdm.devices.traits.ThermostatTemperatureSetpoint"

    async def set_heat(self, heat: float) -> aiohttp.ClientResponse:
        """Change the thermostat Eco mode."""
        data = {
            "command": "sdm.devices.commands.ThermostatTemperatureSetpoint.SetHeat",
            "params": {"heatCelsius": heat},
        }
        return await self.cmd.execute(data)

    async def set_cool(self, cool: float) -> aiohttp.ClientResponse:
        """Change the thermostat Eco mode."""
        data = {
            "command": "sdm.devices.commands.ThermostatTemperatureSetpoint.SetCool",
            "params": {"coolCelsius": cool},
        }
        return await self.cmd.execute(data)

    async def set_range(self, heat: float, cool: float) -> aiohttp.ClientResponse:
        """Change the thermostat Eco mode."""
        data = {
            "command": "sdm.devices.commands.ThermostatTemperatureSetpoint.SetRange",
            "params": {
                "heatCelsius": heat,
                "coolCelsius": cool,
            },
        }
        return await self.cmd.execute(data)



"""Library for transcoding mp4 clips."""

import asyncio.subprocess
import logging
import os

from .exceptions import TranscodeException

_LOGGER = logging.getLogger(__name__)


class Transcoder:
    """A worker that processes mp4 clips."""

    def __init__(self, ffmpeg_binary: str, path_prefix: str) -> None:
        """Initialize transcoder."""
        self._ffmpeg_binary = ffmpeg_binary
        self._path_prefix = path_prefix

    async def transcode_clip(self, input_file: str, output_file: str) -> None:
        """Create a image preview for a thumbnail clip."""
        full_input_file = f"{self._path_prefix}/{input_file}"
        full_output_file = f"{self._path_prefix}/{output_file}"
        if not os.path.exists(full_input_file):
            raise TranscodeException(f"Input file does not exist: {full_input_file}")
        if os.path.exists(full_output_file):
            raise TranscodeException(f"Output file already exists: {full_output_file}")
        cmd = " ".join(
            [
                self._ffmpeg_binary,
                "-y",
                "-i",
                full_input_file,
                "-vf setpts=2.0*PTS",
                "-vf scale=320:-1,setsar=1:1",
                "-r 4",
                "-loop 0",
                full_output_file,
            ]
        )
        proc = await asyncio.create_subprocess_shell(cmd)
        stdout, stderr = await proc.communicate()
        if proc.returncode != 0:
            if stdout:
                _LOGGER.debug(stdout)
            if stderr:
                _LOGGER.debug(stderr)
            raise TranscodeException(
                f"Transcode command failure: {cmd} code: {proc.returncode}"
            )


"""Traits for structures / rooms."""

from __future__ import annotations

from typing import Any, Final, Mapping

try:
    from pydantic.v1 import BaseModel, Field
except ImportError:
    from pydantic import BaseModel, Field  # type: ignore

from .model import TraitModel


class StructureTrait(BaseModel):
    """This trait belongs to any structure for structure-related information."""

    custom_name: str = Field(alias="customName")
    """Name of the structure."""


class InfoTrait(StructureTrait):
    """This trait belongs to any structure for structure-related information."""

    NAME: Final = "sdm.structures.traits.Info"


class RoomInfoTrait(StructureTrait):
    """This trait belongs to any structure for room-related information."""

    NAME: Final = "sdm.structures.traits.RoomInfo"


class Structure(TraitModel):
    """Class that represents a structure object in the Google Nest SDM API."""

    name: str
    """Resource name of the structure e.g. 'enterprises/XYZ/structures/123'."""

    info: InfoTrait | None = Field(alias="sdm.structures.traits.Info")
    room_info: RoomInfoTrait | None = Field(alias="sdm.structures.traits.RoomInfo")

    @staticmethod
    def MakeStructure(raw_data: Mapping[str, Any]) -> Structure:
        """Create a structure with the appropriate traits."""
        return Structure.parse_obj(raw_data)



"""Traits belonging to camera devices."""

from __future__ import annotations

import datetime
import logging
import urllib.parse as urlparse
from abc import ABC, abstractmethod
from collections.abc import Iterable
from dataclasses import dataclass
from enum import Enum
from typing import Final

try:
    from pydantic.v1 import BaseModel, Field, validator
except ImportError:
    from pydantic import BaseModel, Field, validator  # type: ignore

from .event import (
    CameraClipPreviewEvent,
    CameraMotionEvent,
    CameraPersonEvent,
    CameraSoundEvent,
    EventImageContentType,
    EventImageType,
)
from .traits import CommandModel

__all__ = [
    "CameraImageTrait",
    "CameraLiveStreamTrait",
    "CameraEventImageTrait",
    "CameraMotionTrait",
    "CameraPersonTrait",
    "CameraSoundTrait",
    "CameraClipPreviewTrait",
    "Resolution",
    "Stream",
    "StreamUrls",
    "RtspStream",
    "WebRtcStream",
    "StreamingProtocol",
    "EventImage",
]

_LOGGER = logging.getLogger(__name__)

MAX_IMAGE_RESOLUTION = "maxImageResolution"
MAX_VIDEO_RESOLUTION = "maxVideoResolution"
WIDTH = "width"
HEIGHT = "height"
VIDEO_CODECS = "videoCodecs"
AUDIO_CODECS = "audioCodecs"
SUPPORTED_PROTOCOLS = "supportedProtocols"
STREAM_URLS = "streamUrls"
RESULTS = "results"
RTSP_URL = "rtspUrl"
STREAM_EXTENSION_TOKEN = "streamExtensionToken"
STREAM_TOKEN = "streamToken"
URL = "url"
TOKEN = "token"
ANSWER_SDP = "answerSdp"
MEDIA_SESSION_ID = "mediaSessionId"

EVENT_IMAGE_CLIP_PREVIEW = "clip_preview"


@dataclass
class Resolution:
    """Maximum Resolution of an image or stream."""

    width: int | None = None
    height: int | None = None


class CameraImageTrait(BaseModel):
    """This trait belongs to any device that supports taking images."""

    NAME: Final = "sdm.devices.traits.CameraImage"

    max_image_resolution: Resolution | None = Field(alias="maxImageResolution")
    """Maximum resolution of the camera image."""


class Stream(CommandModel, ABC):
    """Base class for streams."""

    expires_at: datetime.datetime = Field(alias="expiresAt")
    """Time at which both streamExtensionToken and streamToken expire."""

    @abstractmethod
    async def extend_stream(self) -> Stream:
        """Extend the lifetime of the stream."""

    @abstractmethod
    async def stop_stream(self) -> None:
        """Invalidate the stream."""


@dataclass
class StreamUrls:
    """Response object for stream urls"""

    rtsp_url: str = Field(alias="rtspUrl")
    """RTSP live stream URL."""


class RtspStream(Stream):
    """Provides access an RTSP live stream URL."""

    stream_urls: StreamUrls = Field(alias="streamUrls")
    """Stream urls to access the live stream."""

    stream_token: str = Field(alias="streamToken")
    """Token to use to access an RTSP live stream."""

    stream_extension_token: str = Field(alias="streamExtensionToken")
    """Token to use to extend access to an RTSP live stream."""

    @property
    def rtsp_stream_url(self) -> str:
        """RTSP live stream URL."""
        return self.stream_urls.rtsp_url

    async def extend_stream(self) -> Stream | RtspStream:
        """Extend the lifetime of the stream."""
        return await self.extend_rtsp_stream()

    async def extend_rtsp_stream(self) -> RtspStream:
        """Request a new RTSP live stream URL access token."""
        data = {
            "command": "sdm.devices.commands.CameraLiveStream.ExtendRtspStream",
            "params": {"streamExtensionToken": self.stream_extension_token},
        }
        response_data = await self.cmd.execute_json(data)
        results = response_data[RESULTS]
        # Update the stream url with the new token
        stream_token = results[STREAM_TOKEN]
        parsed = urlparse.urlparse(self.rtsp_stream_url)
        parsed = parsed._replace(query=f"auth={stream_token}")
        url = urlparse.urlunparse(parsed)
        results[STREAM_URLS] = {}
        results[STREAM_URLS][RTSP_URL] = url
        obj = RtspStream(**results)
        obj._cmd = self.cmd
        return obj

    async def stop_stream(self) -> None:
        """Invalidate the stream."""
        return await self.stop_rtsp_stream()

    async def stop_rtsp_stream(self) -> None:
        """Invalidates a valid RTSP access token and stops the RTSP live stream."""
        data = {
            "command": "sdm.devices.commands.CameraLiveStream.StopRtspStream",
            "params": {"streamExtensionToken": self.stream_extension_token},
        }
        await self.cmd.execute(data)


class WebRtcStream(Stream):
    """Provides access an RTSP live stream URL."""

    answer_sdp: str = Field(alias="answerSdp")
    """An SDP answer to use with the local device displaying the stream."""

    media_session_id: str = Field(alias="mediaSessionId")
    """Media Session ID of the live stream."""

    async def extend_stream(self) -> WebRtcStream:
        """Request a new RTSP live stream URL access token."""
        data = {
            "command": "sdm.devices.commands.CameraLiveStream.ExtendWebRtcStream",
            "params": {MEDIA_SESSION_ID: self.media_session_id},
        }
        response_data = await self.cmd.execute_json(data)
        # Preserve original answerSdp, and merge with response that contains
        # the other fields (expiresAt, and mediaSessionId.
        results = response_data[RESULTS]
        results[ANSWER_SDP] = self.answer_sdp
        obj = WebRtcStream(**results)
        obj._cmd = self.cmd
        return obj

    async def stop_stream(self) -> None:
        """Invalidates a valid RTSP access token and stops the RTSP live stream."""
        data = {
            "command": "sdm.devices.commands.CameraLiveStream.StopWebRtcStream",
            "params": {MEDIA_SESSION_ID: self.media_session_id},
        }
        await self.cmd.execute(data)


class StreamingProtocol(str, Enum):
    """Streaming protocols supported by the device."""

    RTSP = "RTSP"
    WEB_RTC = "WEB_RTC"


class CameraLiveStreamTrait(CommandModel):
    """This trait belongs to any device that supports live streaming."""

    NAME: Final = "sdm.devices.traits.CameraLiveStream"

    max_video_resolution: Resolution = Field(
        alias="maxVideoResolution", default_factory=dict
    )
    """Maximum resolution of the video live stream."""

    video_codecs: list[str] = Field(alias="videoCodecs", default_factory=list)
    """Video codecs supported for the live stream."""

    audio_codecs: list[str] = Field(alias="audioCodecs", default_factory=list)
    """Audio codecs supported for the live stream."""

    supported_protocols: list[StreamingProtocol] = Field(
        alias="supportedProtocols",
        default_factory=list,
    )
    """Streaming protocols supported for the live stream."""

    @validator("supported_protocols", pre=True, always=True)
    def validate_supported_protocols(
        cls, val: Iterable[str] | None
    ) -> list[StreamingProtocol]:
        return [
            StreamingProtocol[x]
            for x in val or []
            if x in StreamingProtocol.__members__
        ] or [StreamingProtocol.RTSP]

    async def generate_rtsp_stream(self) -> RtspStream:
        """Request a token to access an RTSP live stream URL."""
        if StreamingProtocol.RTSP not in self.supported_protocols:
            raise ValueError("Device does not support RTSP stream")
        data = {
            "command": "sdm.devices.commands.CameraLiveStream.GenerateRtspStream",
            "params": {},
        }
        response_data = await self.cmd.execute_json(data)
        results = response_data[RESULTS]
        obj = RtspStream(**results)
        obj._cmd = self.cmd
        return obj

    async def generate_web_rtc_stream(self, offer_sdp: str) -> WebRtcStream:
        """Request a token to access a Web RTC live stream URL."""
        if StreamingProtocol.WEB_RTC not in self.supported_protocols:
            raise ValueError("Device does not support WEB_RTC stream")
        data = {
            "command": "sdm.devices.commands.CameraLiveStream.GenerateWebRtcStream",
            "params": {"offerSdp": offer_sdp},
        }
        response_data = await self.cmd.execute_json(data)
        results = response_data[RESULTS]
        obj = WebRtcStream(**results)
        obj._cmd = self.cmd
        return obj


class EventImage(CommandModel):
    """Provides access to an image in response to an event.

    Use a ?width or ?height query parameters to customize the resolution
    of the downloaded image. Only one of these parameters need to specified.
    The other parameter is scaled automatically according to the camera's
    aspect ratio.

    The token should be added as an HTTP header:
    Authorization: Basic <token>
    """

    event_image_type: EventImageContentType
    """Return the type of event image."""

    url: str | None = Field(default=None)
    """URL to download the camera image from."""

    token: str | None = Field(default=None)
    """Token to use in the HTTP Authorization header when downloading."""

    async def contents(
        self,
        width: int | None = None,
        height: int | None = None,
    ) -> bytes:
        """Download the image bytes."""
        if width:
            fetch_url = f"{self.url}?width={width}"
        elif height:
            fetch_url = f"{self.url}?width={height}"
        else:
            assert self.url
            fetch_url = self.url
        return await self.cmd.fetch_image(fetch_url, basic_auth=self.token)


class CameraEventImageTrait(CommandModel):
    """This trait belongs to any device that generates images from events."""

    NAME: Final = "sdm.devices.traits.CameraEventImage"

    async def generate_image(self, event_id: str) -> EventImage:
        """Provide a URL to download a camera image."""
        data = {
            "command": "sdm.devices.commands.CameraEventImage.GenerateImage",
            "params": {
                "eventId": event_id,
            },
        }
        response_data = await self.cmd.execute_json(data)
        results = response_data[RESULTS]
        img = EventImage(**results, event_image_type=EventImageType.IMAGE)
        img._cmd = self.cmd
        return img


class CameraMotionTrait(BaseModel):
    """For any device that supports motion detection events."""

    NAME: Final = "sdm.devices.traits.CameraMotion"
    EVENT_NAME: Final[str] = CameraMotionEvent.NAME

    class Config:
        extra = "allow"
        arbitrary_types_allowed = True


class CameraPersonTrait(BaseModel):
    """For any device that supports person detection events."""

    NAME: Final = "sdm.devices.traits.CameraPerson"
    EVENT_NAME: Final[str] = CameraPersonEvent.NAME

    class Config:
        extra = "allow"
        arbitrary_types_allowed = True


class CameraSoundTrait(BaseModel):
    """For any device that supports sound detection events."""

    NAME: Final = "sdm.devices.traits.CameraSound"
    EVENT_NAME: Final[str] = CameraSoundEvent.NAME

    class Config:
        extra = "allow"
        arbitrary_types_allowed = True


class CameraClipPreviewTrait(CommandModel):
    """For any device that supports a clip preview."""

    NAME: Final = "sdm.devices.traits.CameraClipPreview"
    EVENT_NAME: Final[str] = CameraClipPreviewEvent.NAME

    async def generate_event_image(self, preview_url: str) -> EventImage | None:
        """Provide a URL to download a camera image from the active event."""
        img = EventImage(url=preview_url, event_image_type=EventImageType.CLIP_PREVIEW)
        img._cmd = self.cmd
        return img



"""Decorator for creating a registry of objects."""

from __future__ import annotations

from typing import Callable, TypeVar

CALLABLE_T = TypeVar("CALLABLE_T", bound=Callable)  # pylint: disable=invalid-name


class Registry(dict):
    """Registry of items."""

    def register(self, name: str | None = None) -> Callable[[CALLABLE_T], CALLABLE_T]:
        """Return decorator to register item with a specific name."""

        def decorator(func: CALLABLE_T) -> CALLABLE_T:
            """Register decorated function."""
            nonlocal name
            if name is None:
                name = func.NAME  # type: ignore
            self[name] = func
            return func

        return decorator



"""Base library for all traits."""

from __future__ import annotations

from typing import Any, Mapping

import aiohttp

try:
    from pydantic.v1 import BaseModel
except ImportError:
    from pydantic import BaseModel  # type: ignore

from .auth import AbstractAuth
from .diagnostics import Diagnostics

DEVICE_TRAITS = "traits"
TRAITS = "traits"


class Command:
    """Base class for executing commands."""

    def __init__(self, device_id: str, auth: AbstractAuth, diagnostics: Diagnostics):
        """Initialize Command."""
        self._device_id = device_id
        self._auth = auth
        self._diagnostics = diagnostics

    async def execute(self, data: Mapping[str, Any]) -> aiohttp.ClientResponse:
        """Run the command."""
        assert self._auth
        cmd = data.get("command", "execute")
        with self._diagnostics.timer(cmd):
            return await self._auth.post(f"{self._device_id}:executeCommand", json=data)

    async def execute_json(self, data: Mapping[str, Any]) -> dict[str, Any]:
        """Run the command and return a json result."""
        assert self._auth
        cmd = data.get("command", "execute")
        with self._diagnostics.timer(cmd):
            return await self._auth.post_json(
                f"{self._device_id}:executeCommand", json=data
            )

    async def fetch_image(self, url: str, basic_auth: str | None = None) -> bytes:
        """Fetch an image at the specified url."""
        headers: dict[str, Any] = {}
        if basic_auth:
            headers = {"Authorization": f"Basic {basic_auth}"}
        with self._diagnostics.timer("fetch_image"):
            resp = await self._auth.get(url, headers=headers)
            return await resp.read()


class CommandModel(BaseModel):
    """Base model that supports commands."""

    _cmd: Command | None = None
    """Helper for executing commands"""

    @property
    def cmd(self) -> Command:
        """Helper for executing commands, used internally by the trait"""
        if not self._cmd:
            raise ValueError("Device trait in invalid state")
        return self._cmd

    class Config:
        extra = "allow"
        arbitrary_types_allowed = True
        fields = {
            "_cmd": {
                "exclude": True,
            },
        }



import pathlib

from setuptools import setup

VERSION = "3.0.2"

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

setup(
    name="google_nest_sdm",
    version=VERSION,
    description="Library for the Google Nest SDM API",
    long_description=README,
    long_description_content_type="text/markdown",
    keywords="google nest sdm camera therostat security doorbell",
    author="Allen Porter",
    author_email="allen@thebends.org",
    url="https://github.com/allenporter/python-google-nest-sdm",
    package_data={"google_nest_sdm": ["py.typed"]},
    packages=["google_nest_sdm"],
    include_package_data=True,
    install_requires=[
        "aiohttp>=3.7.3",
        "google-auth>=1.22.0",
        "google-auth-oauthlib>=0.4.1",
        "google-cloud-pubsub>=2.1.0",
        "requests-oauthlib>=1.3.0",
        "PyYAML>=6.0",
        "pydantic>=1.10.4",
    ],
    python_requires=">=3.10",
    entry_points={
        "console_scripts": [
            "google_nest=google_nest_sdm.google_nest:main",
        ],
    },
    zip_safe=False,
)



"""A device from the Smart Device Management API."""

from __future__ import annotations

import datetime
import logging
from typing import Any, Awaitable, Callable

try:
    from pydantic.v1 import BaseModel, Field
except ImportError:
    from pydantic import BaseModel, Field  # type: ignore

from . import camera_traits, device_traits, doorbell_traits, thermostat_traits
from .auth import AbstractAuth
from .diagnostics import Diagnostics, redact_data
from .event import EventMessage, EventProcessingError
from .event_media import EventMediaManager
from .traits import Command
from .model import TraitModel

_LOGGER = logging.getLogger(__name__)


class ParentRelation(BaseModel):
    """Represents the parent structure/room of the current resource."""

    parent: str
    display_name: str = Field(alias="displayName")


class DeviceTraits(TraitModel):
    """Pydantic model for parsing traits in the Google Nest SDM API."""

    # Device Traits
    connectivity: device_traits.ConnectivityTrait | None = Field(
        alias="sdm.devices.traits.Connectivity", exclude=True
    )
    fan: device_traits.FanTrait | None = Field(
        alias="sdm.devices.traits.Fan", exclude=True
    )
    info: device_traits.InfoTrait | None = Field(
        alias="sdm.devices.traits.Info", exclude=True
    )
    humidity: device_traits.HumidityTrait | None = Field(
        alias="sdm.devices.traits.Humidity", exclude=True
    )
    temperature: device_traits.TemperatureTrait | None = Field(
        alias="sdm.devices.traits.Temperature", exclude=True
    )

    # Thermostat Traits
    thermostat_eco: thermostat_traits.ThermostatEcoTrait | None = Field(
        alias="sdm.devices.traits.ThermostatEco", exclude=True
    )
    thermostat_hvac: thermostat_traits.ThermostatHvacTrait | None = Field(
        alias="sdm.devices.traits.ThermostatHvac", exclude=True
    )
    thermostat_mode: thermostat_traits.ThermostatModeTrait | None = Field(
        alias="sdm.devices.traits.ThermostatMode", exclude=True
    )
    thermostat_temperature_setpoint: thermostat_traits.ThermostatTemperatureSetpointTrait | None = Field(  # noqa: E501
        alias="sdm.devices.traits.ThermostatTemperatureSetpoint", exclude=True
    )

    # Camera Traits
    camera_image: camera_traits.CameraImageTrait | None = Field(
        alias="sdm.devices.traits.CameraImage", exclude=True
    )
    camera_live_stream: camera_traits.CameraLiveStreamTrait | None = Field(
        alias="sdm.devices.traits.CameraLiveStream", exclude=True
    )
    camera_event_image: camera_traits.CameraEventImageTrait | None = Field(
        alias="sdm.devices.traits.CameraEventImage", exclude=True
    )
    camera_motion: camera_traits.CameraMotionTrait | None = Field(
        alias="sdm.devices.traits.CameraMotion", exclude=True
    )
    camera_person: camera_traits.CameraPersonTrait | None = Field(
        alias="sdm.devices.traits.CameraPerson", exclude=True
    )
    camera_sound: camera_traits.CameraSoundTrait | None = Field(
        alias="sdm.devices.traits.CameraSound", exclude=True
    )
    camera_clip_preview: camera_traits.CameraClipPreviewTrait | None = Field(
        alias="sdm.devices.traits.CameraClipPreview", exclude=True
    )

    # Doorbell Traits
    doorbell_chime: doorbell_traits.DoorbellChimeTrait | None = Field(
        alias="sdm.devices.traits.DoorbellChime", exclude=True
    )

    class Config:
        extra = "allow"


class Device(DeviceTraits):
    """Class that represents a device object in the Google Nest SDM API."""

    name: str
    """Resource name of the device such as 'enterprises/XYZ/devices/123'."""

    type: str | None
    """Type of device for display purposes.

    The device type should not be used to deduce or infer functionality of
    the actual device it is assigned to. Instead, use the returned traits for
    the device.
    """

    relations: list[ParentRelation] = Field(
        alias="parentRelations", default_factory=list
    )
    """Represents the parent structure or room of the device."""

    def __init__(self, raw_data: dict[str, Any], auth: AbstractAuth) -> None:
        """Initialize a device."""
        # Hack for incorrect nest API response values
        if (type := raw_data.get("type")) and type == "sdm.devices.types.DOORBELL":
            if "traits" not in raw_data:
                raw_data["traits"] = {}
            raw_data["traits"][doorbell_traits.DoorbellChimeTrait.NAME] = {}
        super().__init__(**raw_data)
        self._auth = auth
        self._diagnostics = Diagnostics()
        self._cmd = Command(raw_data["name"], auth, self._diagnostics.subkey("command"))
        for trait in self.traits.values():
            if hasattr(trait, "_cmd"):
                trait._cmd = self._cmd

        event_traits = {
            trait.EVENT_NAME
            for trait in self.traits.values()
            if hasattr(trait, "EVENT_NAME")
        }
        self._event_media_manager = EventMediaManager(
            self.name,
            self.traits,
            event_traits,
            diagnostics=self._diagnostics.subkey("event_media"),
        )
        self._callbacks: list[Callable[[EventMessage], Awaitable[None]]] = []
        self._trait_event_ts: dict[str, datetime.datetime] = {}

    @staticmethod
    def MakeDevice(raw_data: dict[str, Any], auth: AbstractAuth) -> Device:
        """Create a device with the appropriate traits."""
        return Device(raw_data, auth)

    def add_update_listener(self, target: Callable[[], None]) -> Callable[[], None]:
        """Register a simple event listener notified on updates.

        The return value is a callable that will unregister the callback.
        """

        async def handle_event(event_message: EventMessage) -> None:
            target()

        return self.add_event_callback(handle_event)

    def add_event_callback(
        self, target: Callable[[EventMessage], Awaitable[None]]
    ) -> Callable[[], None]:
        """Register an event callback for updates to this device.

        The return value is a callable that will unregister the callback.
        """
        self._callbacks.append(target)

        def remove_callback() -> None:
            """Remove the event_callback."""
            self._callbacks.remove(target)

        return remove_callback

    async def async_handle_event(self, event_message: EventMessage) -> None:
        """Process an event from the pubsub subscriber."""
        _LOGGER.debug(
            "Processing update %s @ %s", event_message.event_id, event_message.timestamp
        )
        if not event_message.resource_update_name:
            raise EventProcessingError("Event was not resource update event")
        if self.name != event_message.resource_update_name:
            raise EventProcessingError(
                f"Mismatch {self.name} != {event_message.resource_update_name}"
            )
        self._async_handle_traits(event_message)
        await self._event_media_manager.async_handle_events(event_message)
        for callback in self._callbacks:
            await callback(event_message)

    def _async_handle_traits(self, event_message: EventMessage) -> None:
        traits = event_message.resource_update_traits
        if not traits:
            return
        _LOGGER.debug("Trait update %s", traits)
        # Parse the traits using a separate pydantic object, then overwrite
        # each present field with an updated copy of the original trait with
        # the new fields merged in.
        parsed_traits = DeviceTraits.parse_obj({"traits": traits})
        for field in parsed_traits.__fields__.values():
            if not (new := getattr(parsed_traits, field.name)) or not isinstance(
                new, BaseModel
            ):
                continue
            # Discard updates to traits that are newer than the update
            if (
                self._trait_event_ts
                and (ts := self._trait_event_ts.get(field.name))
                and ts > event_message.timestamp
            ):
                _LOGGER.debug("Discarding stale update (%s)", event_message.timestamp)
                continue

            # Only merge updates into existing models
            if not (existing := getattr(self, field.name)) or not isinstance(
                existing, BaseModel
            ):
                continue
            obj = existing.copy(update=new.dict())
            setattr(self, field.name, obj)
            self._trait_event_ts[field.name] = event_message.timestamp

    @property
    def event_media_manager(self) -> EventMediaManager:
        return self._event_media_manager

    @property
    def parent_relations(self) -> dict:
        """Room or structure for the device."""
        return {relation.parent: relation.display_name for relation in self.relations}

    def delete_relation(self, parent: str) -> None:
        """Remove a device relationship with the parent."""
        self.relations = [
            relation for relation in self.relations if relation.parent != parent
        ]

    def create_relation(self, relation: ParentRelation) -> None:
        """Add a new device relation."""
        self.relations.append(relation)

    def get_diagnostics(self) -> dict[str, Any]:
        return {
            "data": redact_data(self.raw_data),
            **self._diagnostics.as_dict(),
        }

    _EXCLUDE_FIELDS = (
        set({"_auth", "_callbacks", "_cmd", "_diagnostics", "_event_media_manager"})
        | TraitModel._EXCLUDE_FIELDS
    )

    class Config:
        arbitrary_types_allowed = True
        extra = "allow"
